# %%

import numpy as np
import matplotlib.pyplot as plt
from .core.utility import Parameters
from .core.tessellation import Tessellation



class Cpab:
    """ Core class for this library. This class contains all the information
        about the tesselation, transformation etc. The user is not meant to
        use anything else than this specific class.
        
    Arguments:
        tess_size: list, with the number of cells in each dimension
        
        backend: string, computational backend to use. Choose between 
            "numpy" (default), or "pytorch"
        
        device: string, either "cpu" (default) or "gpu". For the numpy backend
            only the "cpu" option is valid
        
        zero_boundary: bool, determines is the velocity at the boundary is zero 
        
        basis: string, constrain basis to use. Choose between 
            "rref" (default), "svd", or "sparse"
        
    Methods:
        @uniform_meshgrid
        @sample_transformation
        @sample_transformation_with_prior
        @identity
        @transform_grid
        @interpolate
        @transform_data
        @get_velocity
        @visualize_velocity
        @visualize_deformgrid
        @visualize_tesselation
    """

    def __init__(self, tess_size, backend="numpy", device="cpu", zero_boundary=True, basis="rref"):
        # Check input
        self._check_input(tess_size, backend, device, zero_boundary, basis)

        # Parameters
        self.params = Parameters()
        self.params.nc = tess_size
        self.params.zero_boundary = zero_boundary
        self.params.xmin = 0
        self.params.xmax = 1
        self.params.nSteps1 = 10
        self.params.nSteps2 = 5
        self.params.precomputed = False
        self.params.use_slow = False
        self.params.basis = basis

        # Initialize tesselation
        self.tess = Tessellation(
            self.params.nc,
            self.params.xmin,
            self.params.xmax,
            self.params.zero_boundary,
            basis=self.params.basis,
        )

        # Extract parameters from tesselation
        self.params.B = self.tess.B
        self.params.D, self.params.d = self.tess.D, self.tess.d

        self.backend_name = backend

        # Load backend and set device
        self.device = device.lower()
        self.backend_name = backend
        if self.backend_name == "numpy":
            from .backend.numpy import functions as backend
        # elif self.backend_name == "numba":
        #     pass
        #     from .tensorflow import functions as backend
        elif self.backend_name == "pytorch":
            from .backend.pytorch import functions as backend
            self.params.B = backend.to(self.params.B, device=self.device)
            self.params.B = self.params.B.contiguous()
        self.backend = backend

        # Assert that we have a recent version of the backend
        self.backend.assert_version()

    def uniform_meshgrid(self, n_points):
        """ Constructs a meshgrid 
        Arguments:
            n_points: number of points
        Output:
            grid: vector of points
        """
        return self.backend.uniform_meshgrid(
            self.params.xmin, self.params.xmax, n_points, self.device
        )

    def covariance_cpa(self, length_scale=0.1, output_variance=1):
        """ Function for sampling smooth transformations. The smoothness is determined
            by the distance between cell centers. The closer the cells are to each other,
            the more the cell parameters should correlate -> smooth transistion in
            parameters. The covariance in the D-space is calculated using the
            squared exponential kernel.
                
        Arguments:
            n_sample: integer, number of transformation to sample
            mean: [d,] vector, mean of multivariate gaussian
            length_scale: float>0, determines how fast the covariance declines 
                between the cells 
            output_variance: float>0, determines the overall variance from the mean
        Output:
            samples: [n_sample, d] matrix. Each row is a independent sample from
                a multivariate gaussian
        """

        centers = self.backend.to(self.tess.cell_centers(), device=self.device)
        dist = self.backend.pdist(centers)

        cov_init = self.backend.ones((self.params.D, self.params.D))
        cov_init = self.backend.to(cov_init, device=self.device)*100*self.backend.max(dist)
        cov_init[::2,::2] = dist
        cov_init[1::2,1::2] = dist

        # Squared exponential kernel + Smoothness priors on CPA
        cov_pa = output_variance**2 * self.backend.exp(-(cov_init / (2*length_scale**2)))       

        B = self.params.B
        cov_cpa = self.backend.matmul(self.backend.matmul(B.T, cov_pa), B)

        return self.backend.to(cov_cpa, device=self.device)

    def sample_transformation(self, n_sample, mean=None, cov=None):
        """ Method for sampling transformation from simply multivariate gaussian
            As default the method will sample from a standard normal
        Arguments:
            n_sample: integer, number of transformations to sample
            mean: [d,] vector, mean of multivariate gaussian
            cov: [d,d] matrix, covariance of multivariate gaussian
        Output:
            samples: [n_sample, d] matrix. Each row is a independent sample from
                a multivariate gaussian
        """
        if mean is not None:
            self._check_type(mean)
            self._check_device(mean)
        if cov is not None:
            self._check_type(cov)
            self._check_device(cov)

        samples = self.backend.sample_transformation(
            self.params.d, n_sample, mean, cov, self.device
        )

        return self.backend.to(samples, device=self.device)

    def sample_transformation_with_prior(
        self, n_sample, mean=None, length_scale=0.1, output_variance=1
    ):
        """ Function for sampling smooth transformations. The smoothness is determined
            by the distance between cell centers. The closer the cells are to each other,
            the more the cell parameters should correlate -> smooth transistion in
            parameters. The covariance in the D-space is calculated using the
            squared exponential kernel.
                
        Arguments:
            n_sample: integer, number of transformation to sample
            mean: [d,] vector, mean of multivariate gaussian
            length_scale: float>0, determines how fast the covariance declines 
                between the cells 
            output_variance: float>0, determines the overall variance from the mean
        Output:
            samples: [n_sample, d] matrix. Each row is a independent sample from
                a multivariate gaussian
        """

        if mean is not None:
            self._check_type(mean)
            self._check_device(mean)


        centers = self.backend.to(self.tess.cell_centers(), device=self.device)
        dist = self.backend.pdist(centers)

        cov_init = self.backend.ones((self.params.D, self.params.D))
        cov_init = self.backend.to(cov_init, device=self.device)*100*self.backend.max(dist)
        cov_init[::2,::2] = dist
        cov_init[1::2,1::2] = dist

        # Squared exponential kernel + Smoothness priors on CPA
        cov_pa = output_variance**2 * self.backend.exp(-(cov_init / (2*length_scale**2)))       

        B = self.params.B
        cov_cpa = self.backend.matmul(self.backend.matmul(B.T, cov_pa), B)

        samples = self.backend.sample_transformation(
            self.params.d, n_sample, mean, cov_cpa, self.device
        )
        return self.backend.to(samples, device=self.device)

    def identity(self, n_sample=1, epsilon=0):
        """ Method for getting the parameters for the identity 
            transformation (vector of zeros) 
        Arguments:
            n_sample: integer, number of transformations to sample
            epsilon: float>0, small number to add to the identity transformation
                for stability during training
        Output:
            samples: [n_sample, d] matrix. Each row is a sample    
        """
        return self.backend.identity(self.params.d, n_sample, epsilon, self.device)

    def transform_grid(self, grid, theta, method=None, time=1.0):
        """ Main method of the class. Integrates the grid using the parametrization
            in theta.
        Arguments:
            grid: [n_points] vector or [n_batch, n_points] tensor i.e.
                either a single grid for all theta values, or a grid for each theta
                value
            theta: [n_batch, d] matrix,
            method: one of {"closed_form", "numeric"}
        Output:
            transformed_grid: [n_batch, n_points] tensor, with the transformed
                grid. The slice transformed_grid[i] corresponds to the grid being
                transformed by theta[i]
        """
        self._check_type(grid)
        self._check_device(grid)
        self._check_type(theta)
        self._check_device(theta)
        transformed_grid = self.backend.transformer(grid, theta, self.params, method, time)
        return transformed_grid

    def transform_grid_ss(self, grid, theta, method=None, time=1.0, N=0):
        """ Main method of the class. Integrates the grid using the parametrization
            in theta.
        Arguments:
            grid: [n_points] vector or [n_batch, n_points] tensor i.e.
                either a single grid for all theta values, or a grid for each theta
                value
            theta: [n_batch, d] matrix,
            method: one of {"closed_form", "numeric"}
        Output:
            transformed_grid: [n_batch, n_points] tensor, with the transformed
                grid. The slice transformed_grid[i] corresponds to the grid being
                transformed by theta[i]
        """
        self._check_type(grid)
        self._check_device(grid)
        self._check_type(theta)
        self._check_device(theta)
        # time = time / 2**N
        transformed_grid = self.backend.transformer(grid, theta, self.params, method, time)
        for i in range(N):
            transformed_grid = self.backend.interpolate_grid(transformed_grid, self.params)
        return transformed_grid

    def gradient_grid(self, grid, theta, method=None, time=1.0):
        """ Integrates and return the gradient of the transformation
        using the parametrization in theta.
        Arguments:
            grid: [n_points] vector or [n_batch, n_points] tensor i.e.
                either a single grid for all theta values, or a grid for each theta
                value
            theta: [n_batch, d] matrix,
        Output:
            transformed_grid: [n_batch, n_points] tensor, with the transformed
                grid. The slice transformed_grid[i] corresponds to the grid being
                transformed by theta[i]
            gradient_grid: [n_batch, n_points, d] tensor, with the gradient grid.
                The slice gradient_grid[i, :, j] corresponds to the gradient of grid being
                transformed by theta[i] and with respect to the parameter theta[i,j]
        """
        self._check_type(grid)
        self._check_device(grid)
        self._check_type(theta)
        self._check_device(theta)
        transformed_grid = self.backend.gradient(grid, theta, self.params, method, time)
        return transformed_grid

    def interpolate(self, data, grid, outsize):
        """ Linear interpolation method
        Arguments:
            data: [n_batch, *data_shape] tensor, with input data. The format of
                the data_shape depends on the backend that is being used. 
                In tensorflow and numpy: [n_batch, number_of_features, n_channels]
                In pytorch: [n_batch, n_channels, number_of_features]
            grid: [n_batch, n_points] tensor with grid points that are 
                used to interpolate the data
            outsize: number of points in the output
        Output:
            interpolated: [n_batch, outsize, n_channels] tensor with the interpolated data
        """
        self._check_type(data)
        self._check_device(data)
        self._check_type(grid)
        self._check_device(grid)
        return self.backend.interpolate(data, grid, outsize)

    def transform_data(self, data, theta, outsize, method=None, time=1.0):
        """ Combination of the transform_grid and interpolate methods for easy
            transformation of data.
        Arguments:
            data: [n_batch, *data_shape] tensor, with input data. The format of
                the data_shape depends on the backend that is being used. 
                In tensorflow and numpy: [n_batch, number_of_features, n_channels]
                In pytorch: [n_batch, n_channels, number_of_features]
            theta: [n_batch, d] matrix with transformation parameters. Each row
                correspond to a transformation.
            outsize: number of points that is transformed and interpolated
        Output:
            data_t: [n_batch, outsize, n_channels] tensor, transformed and interpolated data
        """

        self._check_type(data)
        self._check_device(data)
        self._check_type(theta)
        self._check_device(theta)
        assert (
            data.shape[0] == theta.shape[0]
        ), """Batch sizes should be the same on arguments data and theta"""
        grid = self.uniform_meshgrid(outsize)
        grid_t = self.transform_grid(grid, theta, method, time)
        data_t = self.interpolate(data, grid_t, outsize)
        return data_t

    def transform_data_ss(self, data, theta, outsize, method=None, time=1.0, N=0):
        """ Combination of the transform_grid and interpolate methods for easy
            transformation of data.
        Arguments:
            data: [n_batch, *data_shape] tensor, with input data. The format of
                the data_shape depends on the backend that is being used. 
                In tensorflow and numpy: [n_batch, number_of_features, n_channels]
                In pytorch: [n_batch, n_channels, number_of_features]
            theta: [n_batch, d] matrix with transformation parameters. Each row
                correspond to a transformation.
            outsize: number of points that is transformed and interpolated
        Output:
            data_t: [n_batch, outsize, n_channels] tensor, transformed and interpolated data
        """

        self._check_type(data)
        self._check_device(data)
        self._check_type(theta)
        self._check_device(theta)
        assert (
            data.shape[0] == theta.shape[0]
        ), """Batch sizes should be the same on arguments data and theta"""
        grid = self.uniform_meshgrid(outsize)
        grid_t = self.transform_grid_ss(grid, theta, method, time, N)
        data_t = self.interpolate(data, grid_t, outsize)
        return data_t

    def calc_velocity(self, grid, theta):
        """ For each point in grid, calculate the velocity of the point based
            on the parametrization in theta
        Arguments:
            grid: [n_points] vector
            theta: [1, d] single parametrization vector
        Output:    
            v: [n_points] vector with velocity vectors for each point
        """
        self._check_type(grid)
        self._check_device(grid)
        self._check_type(theta)
        self._check_device(theta)
        v = self.backend.calc_velocity(grid, theta, self.params)
        return v

    def visualize_velocity(self, theta, n_points=None, fig=None):
        """ Utility function that helps visualize the vectorfield for a specific
            parametrization vector theta 
        Arguments:    
            theta: [n_batch, d] single parametrization vector
            n_points: number of points to plot
            fig: matplotlib figure handle
        Output:
            plot: handle to lines plot
        """
        self._check_type(theta)
        if fig is None:
            fig = plt.figure()

        if n_points is None:
            n_points = self.params.nc+1

        # Calculate vectorfield and convert to numpy
        grid = self.uniform_meshgrid(n_points)
        v = self.calc_velocity(grid, theta)

        grid = self.backend.tonumpy(grid)
        v = self.backend.tonumpy(v)

        # Plot
        ax = fig.add_subplot(1, 1, 1)
        ax.axhline(color="black", ls="dashed")
        alpha = max(0.01, 1/np.sqrt(len(theta)))
        ax.plot(grid, v.T, color="blue", alpha=alpha)
        ax.plot(grid, v.T, marker='o', linestyle="None", color="orange", alpha=alpha)
        ax.set_xlim(self.params.xmin, self.params.xmax)
        ax.set_title("Velocity Field " + r'$v(x)$')
        ax.set_xlabel(r'$x$', rotation='horizontal')
        ax.set_ylabel(r'$v(x)$', rotation='horizontal')
        ax.grid(alpha=0.3)
        return ax

    def visualize_deformgrid(self, theta, method=None, time=1.0, n_points=100, fig=None):
        """ Utility function that helps visualize a deformation
        Arguments:
            theta: [n_batch, d] single parametrization vector
            n_points: int, number of points
            fig: matplotlib figure handle
        Output:
            plot: handle to lineplot
        """
        pass
        if fig is None:
            fig = plt.figure()

        # Transform grid and convert to numpy
        grid = self.uniform_meshgrid(n_points)
        grid_t = self.transform_grid(grid, theta, method, time)
        grid = self.backend.tonumpy(grid)
        grid_t = self.backend.tonumpy(grid_t)

        # Plot
        ax = fig.add_subplot(1, 1, 1)
        ax.axline((0, 0), (1, 1), color="black", ls="dashed")
        alpha = max(0.01, 1/np.sqrt(len(theta))) 
        ax.plot(grid, grid_t.T, c="blue", alpha=alpha)
        ax.set_title("Grid Deformation " + r'$\phi(x,t)$')
        ax.set_xlabel(r'$x$', rotation='horizontal')
        ax.set_ylabel(r'$\phi(x,t)$', rotation='horizontal')
        ax.grid(alpha=0.3)
        
        return ax

    def visualize_tesselation(self, n_points=50, fig=None):
        """ Utility function that helps visualize the tesselation.
        Arguments:
            n_points: number of points in each dimension
            fig: matplotlib figure handle
        Output:
            plot: handle to tesselation plot
        """
        if fig is None:
            fig = plt.figure()

        grid = self.uniform_meshgrid(n_points)

        # Find cellindex and convert to numpy
        idx = self.backend.get_cell(grid, self.params)
        idx = self.backend.tonumpy(idx)
        grid = self.backend.tonumpy(grid)

        # Plot
        ax = fig.add_subplot(1, 1, 1)
        plot = ax.scatter(grid.flatten(), idx, c=idx)
        ax.set_title("Tesselation " + r'$N_\mathcal{P}=$' + str(self.params.nc))
        ax.set_xlabel(r'$x$')
        ax.set_ylabel("Cell Index")
        ax.set_yticks(np.arange(np.max(idx)+1))
        ax.grid(alpha=0.3)
        return plot
        

    def visualize_gradient(self, theta, method=None, time=1.0, n_points=100, fig=None):
        """ Utility function that helps visualize the gradient
        Arguments:
            theta: [n_batch, d] single parametrization vector
            n_points: int, number of points
            fig: matplotlib figure handle
        Output:
            plot: handle to lineplot
        """
        if fig is None:
            fig = plt.figure()

        # Gradient grid and convert to numpy
        grid = self.uniform_meshgrid(n_points)
        grad = self.gradient_grid(grid, theta, method, time)
        grid = self.backend.tonumpy(grid)
        grad = self.backend.tonumpy(grad)

        # Plot only gradient
        ax = fig.add_subplot(1, 1, 1)
        ax.axhline(color="black", ls="dashed")
        for g in grad:
            ax.plot(grid, g)
        ax.set_title("Gradient " + r'$\dfrac{\partial \phi(x,t)}{\partial \theta}$')
        ax.set_xlabel(r'$x$', rotation='horizontal')
        ax.set_ylabel(r'$\dfrac{\partial \phi(x,t)}{\partial \theta}$', rotation='horizontal', labelpad=10)
        ax.grid(alpha=0.3)
        return ax

    def visualize_deformdata(self, data, theta, method=None, outsize=100, target=None, fig=None):
        """ Utility function that helps visualize a deformation
        Arguments:
            theta: [n_batch, d] single parametrization vector
            n_points: int, number of points
            fig: matplotlib figure handle
        Output:
            plot: handle to lineplot
        """
        pass
        if fig is None:
            fig = plt.figure()

        # Transform grid and convert to numpy
        data_t = self.transform_data(data, theta, outsize, method)
        data_t = self.backend.tonumpy(data_t)

        batch_size, width, channels = data.shape
        x = np.linspace(0,1,width)
        xt = np.linspace(0,1,outsize)
        alpha = max(0.01, 1/np.sqrt(batch_size))

        # Plot
        plt.suptitle("Data Deformation with " + r'$\phi(x,t)$')
        for i in range(channels):
            ax = fig.add_subplot(channels, 1, i+1)
            ax.plot(x, data[:,:,i].T, c="red", ls="dashed", alpha=alpha, label="Data")
            if target is not None:
                ax.plot(xt, target[:,:,i].T, c="green", ls="dashdot", alpha=alpha, label="Target")
            ax.plot(xt, data_t[:,:,i].T, alpha=alpha, label="Transformed")
            ax.set_title("Ch " + str(i), rotation=-90, loc='right', y=0.5, ha="left", va="center")
            ax.set_ylabel(r"$x'$", rotation='horizontal')
            ax.grid(alpha=0.3)
            if i+1 < channels:
                ax.set_xticklabels([])
            else:
                ax.set_xlabel(r"$x$", rotation='horizontal')
        
        return ax


    def _check_input(self, tess_size, backend, device, zero_boundary, basis):
        """ Utility function used to check the input to the class.
            Not meant to be called by the user. """
        assert tess_size > 1, """tess size must be > 1"""
        assert backend in [
            "numpy",
            "pytorch",
        ], """Unknown backend, choose between 'numpy' or 'pytorch' """
        assert device in [
            "cpu",
            "gpu",
        ], """Unknown device, choose between 'cpu' or 'gpu' """
        if backend == "numpy":
            assert device == "cpu", """Cannot use gpu with numpy backend """
        assert (
            type(zero_boundary) == bool
        ), """Argument zero_boundary must be True or False"""
        assert basis in [
            "svd",
            "rref",
            "sparse",
            "qr"
        ], """Unknown basis, choose between 'svd', 'rref' 'qr', or 'sparse' """
    def _check_type(self, x):
        """ Assert that the type of x is compatible with the class i.e
                numpy backend expects np.array
                pytorch backend expects torch.tensor
                tensorflow backend expects tf.tensor
        """
        assert isinstance(
            x, self.backend.backend_type()
        ), """ Input has type {0} but expected type {1} """.format(
            type(x), self.backend.backend_type()
        )

    def _check_device(self, x):
        """ Assert that x is on the same device (cpu or gpu) as the class """
        assert self.backend.check_device(
            x, self.device
        ), """Input is placed on 
            device {0} but the class expects it to be on device {1}""".format(
            str(x.device), self.device
        )

    def __repr__(self):
        output = """
        CPAB transformer class. 
            Parameters:
                Tesselation size:           {0}
                Theta size:                 {1}
                Domain lower bound:         {2}
                Domain upper bound:         {3}
                Zero Boundary:              {4}
            Backend:                        {5}
        """.format(
            self.params.nc,
            self.params.d,
            self.params.xmin,
            self.params.xmax,
            self.params.zero_boundary,
            self.backend_name,
        )
        return output



# %% SETUP

import numpy as np

eps = np.finfo(np.float32).eps
np.seterr(divide="ignore", invalid="ignore")

# %% BATCH EFFECT

def batch_effect(x, theta):
    if x.ndim == 1:
        n_batch = theta.shape[0]
        n_points = x.shape[-1]
        x = np.broadcast_to(x, (n_batch, n_points))#.flatten()
    return x.flatten()

# %% FUNCTIONS

def get_affine(x, theta, params):
    if params.precomputed:
        return params.A, params.r
    else:
        n_batch = theta.shape[0]
        n_points = x.shape[-1]

        # r = np.broadcast_to(np.arange(n_batch), [n_points, n_batch]).T
        # NOTE: here we suppose batch effect has been already executed
        r = np.arange(n_batch).repeat(n_points / n_batch)

        A = params.B.dot(theta.T).T.reshape(n_batch, -1, 2)

        return A, r

def precompute_affine(x, theta, params):
    params = params.copy()
    params.precomputed = False
    params.A, params.r = get_affine(x, theta, params)
    params.precomputed = True
    return params

def right_boundary(c, params):
    xmin, xmax, nc = params.xmin, params.xmax, params.nc
    return xmin + (c + 1) * (xmax - xmin) / nc + eps

def left_boundary(c, params):
    xmin, xmax, nc = params.xmin, params.xmax, params.nc
    return xmin + c * (xmax - xmin) / nc - eps

def get_cell(x, params):
    xmin, xmax, nc = params.xmin, params.xmax, params.nc

    c = np.floor((x - xmin) / (xmax - xmin) * nc)
    c = np.clip(c, 0, nc - 1).astype(np.int32)
    return c

def get_velocity(x, theta, params):
    A, r = get_affine(x, theta, params)
    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]
    return a * x + b

def get_psi(x, t, theta, params):
    A, r = get_affine(x, theta, params)
    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]

    cond = a == 0
    x1 = x + t * b
    eta = np.exp(t * a)
    x2 = eta * x + (b / a) * (eta-1.0)
    # x2 = np.exp(t * a) * (x + (b / a)) - (b / a)
    psi = np.where(cond, x1, x2)
    return psi

def get_hit_time(x, theta, params):
    A, r = get_affine(x, theta, params)

    c = get_cell(x, params)
    v = get_velocity(x, theta, params)
    xc = np.where(v >= 0, right_boundary(c, params), left_boundary(c, params))

    a = A[r, c, 0]
    b = A[r, c, 1]

    cond = a == 0
    x1 = (xc - x) / b
    x2 = np.log((xc + b / a) / (x + b / a)) / a
    thit = np.where(cond, x1, x2)
    return thit

def get_phi_numeric(x, t, theta, params):
    nSteps2 = params.nSteps2

    yn = x
    deltaT = t / nSteps2
    for j in range(nSteps2):
        c = get_cell(yn, params)
        midpoint = yn + deltaT / 2 * get_velocity(yn, theta, params)
        c = get_cell(midpoint, params)
        yn = yn + deltaT * get_velocity(midpoint, theta, params)
    return yn


# %% INTEGRATION


def integrate_numeric(x, theta, params, time=1.0):
    # setup
    x = batch_effect(x, theta)
    n_batch = theta.shape[0]
    t = time
    params = precompute_affine(x, theta, params)

    # computation
    xPrev = x
    nSteps1 = params.nSteps1
    deltaT = t / nSteps1
    c = get_cell(xPrev, params)
    for j in range(nSteps1):
        xTemp = get_psi(xPrev, deltaT, theta, params)
        cTemp = get_cell(xTemp, params)
        xNum = get_phi_numeric(xPrev, deltaT, theta, params)
        xPrev = np.where(c == cTemp, xTemp, xNum)
        c = get_cell(xPrev, params)
    return xPrev.reshape((n_batch, -1))

def integrate_closed_form(x, theta, params, time=1.0):
    # setup
    x = batch_effect(x, theta)
    t = np.ones_like(x)*time
    params = precompute_affine(x, theta, params)
    n_batch = theta.shape[0]

    # computation
    phi = np.empty_like(x)
    done = np.full_like(x, False, dtype=bool)
    c = get_cell(x, params)

    cont = 0
    while True:
        left = left_boundary(c, params)
        right = right_boundary(c, params)
        v = get_velocity(x, theta, params)
        psi = get_psi(x, t, theta, params)

        cond1 = np.logical_and(left <= psi, psi <= right)
        cond2 = np.logical_and(v >= 0, c == params.nc-1)
        cond3 = np.logical_and(v <= 0, c == 0)
        valid = np.any((cond1, cond2, cond3), axis=0)

        phi[~done] = psi
        done[~done] = valid
        if np.alltrue(valid):
            return phi.reshape((n_batch, -1))

        x, t, params.r = x[~valid], t[~valid], params.r[~valid]
        t -= get_hit_time(x, theta, params)
        x = np.clip(psi, left, right)[~valid]
        c = np.where(v >= 0, c+1, c-1)[~valid]

        cont += 1
        if cont > params.nc:
            raise BaseException
    return None

def integrate_closed_form_trace(x, theta, params, time=1.0):

    x = batch_effect(x, theta)
    t = np.ones_like(x)*time
    params = precompute_affine(x, theta, params)

    result = np.empty((*x.shape, 3))
    done = np.full_like(x, False, dtype=bool)
    
    c = get_cell(x, params)
    cont = 0
    while True:
        left = left_boundary(c, params)
        right = right_boundary(c, params)
        v = get_velocity(x, theta, params)
        psi = get_psi(x, t, theta, params)

        cond1 = np.logical_and(left <= psi, psi <= right)
        cond2 = np.logical_and(v >= 0, c == params.nc-1)
        cond3 = np.logical_and(v <= 0, c == 0)
        valid = np.any((cond1, cond2, cond3), axis=0)

        result[~done] = np.array([psi, t, c]).T
        done[~done] = valid
        if np.alltrue(valid):
            return result

        x, t, params.r = x[~valid], t[~valid], params.r[~valid]
        t -= get_hit_time(x, theta, params)
        x = np.clip(psi, left, right)[~valid]
        c = np.where(v >= 0, c+1, c-1)[~valid]

        cont += 1
        if cont > params.nc:
            raise BaseException
    return None

# %% DERIVATIVE

def derivative_numeric(x, theta, params, time=1.0, h=1e-3):
    # setup
    n_points = x.shape[-1]
    n_batch = theta.shape[0]
    d = theta.shape[1]

    # computation
    der = np.empty((n_batch, n_points, d))

    phi_1 = integrate_numeric(x, theta, params, time)
    for k in range(d):
        theta2 = theta.copy()
        theta2[:,k] += h
        phi_2 = integrate_numeric(x, theta2, params, time)

        der[:,:,k] = (phi_2 - phi_1)/h

    return phi_1, der

def derivative_closed_form(x, theta, params, time=1.0):
    # setup
    n_points = x.shape[-1]
    n_batch = theta.shape[0]
    d = theta.shape[1]

    # computation
    result = integrate_closed_form_trace(x, theta, params, time)
    phi = result[:,0].reshape((n_batch, -1))
    tm = result[:,1]
    cm = result[:,2]

    # setup
    x = batch_effect(x, theta)
    params = precompute_affine(x, theta, params)

    der = np.empty((n_batch, n_points, d))
    for k in range(d):
        dthit_dtheta_cum = np.zeros_like(x)
 
        xm = x.copy()
        c = get_cell(x, params)
        while True:
            valid = c == cm
            if np.alltrue(valid):
                break
            step = np.sign(cm-c)
            dthit_dtheta_cum[~valid] -= derivative_thit_theta(xm, theta, k, params)[~valid]
            xm[~valid] = np.where(step == 1, right_boundary(c, params), left_boundary(c, params))[~valid]
            c = c + step

        dpsi_dtheta = derivative_psi_theta(xm, tm, theta, k, params)
        dpsi_dtime = derivative_phi_time(xm, tm, theta, k, params)
        dphi_dtheta = dpsi_dtheta + dpsi_dtime * dthit_dtheta_cum
        der[:,:,k] = dphi_dtheta.reshape(n_batch, n_points)
    
    return phi, der

def derivative_psi_theta(x, t, theta, k, params):
    A, r = get_affine(x, theta, params)
    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]

    ak = params.B[2 * c, k]
    bk = params.B[2 * c + 1, k]

    cond = a == 0
    d1 = t * (x * ak + bk)
    d2 = (
        ak * t * np.exp(a * t) * (x + b / a)
        + (np.exp(t * a) - 1) * (bk * a - ak * b) / a ** 2
    )
    dpsi_dtheta = np.where(cond, d1, d2)
    return dpsi_dtheta

def derivative_phi_time(x, t, theta, k, params):
    A, r = get_affine(x, theta, params)
    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]

    cond = a == 0
    d1 = b
    d2 = np.exp(t * a) * (a * x + b)
    dpsi_dtime = np.where(cond, d1, d2)
    return dpsi_dtime

def derivative_thit_theta(x, theta, k, params):
    A, r = get_affine(x, theta, params)
    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]

    ak = params.B[2 * c, k]
    bk = params.B[2 * c + 1, k]

    v = get_velocity(x, theta, params)

    xc = np.where(v >= 0, right_boundary(c, params), left_boundary(c, params))

    cond = a == 0
    d1 = (x - xc) * bk / b ** 2
    d2 = -ak * np.log((a * xc + b) / (a * x + b)) / a ** 2
    d3 = (x - xc) * (bk * a - ak * b) / (a * (a * x + b) * (a * xc + b))
    dthit_dtheta = np.where(cond, d1, d2 + d3)
    return dthit_dtheta
